/*
 * Name:    main.cpp
 *
 * Purpose: This is the main routine for the MX CA server.
 *
 *-----------------------------------------------------------------------
 *
 * Copyright 2004, 2008-2009 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <iostream>

#include "mxcaServer.h"

#include "fdManager.h"

extern "C" {
#include "mx_unistd.h"		/* For getopt() */
}

int
main( int argc, char **argv )
{
	mxcaServer *server;
	MX_RECORD *mx_record_list;
	MX_LIST_HEAD *mx_list_head;
	MX_PIPE *mx_callback_pipe;
	MX_CALLBACK_MESSAGE *mx_callback_message;
	MX_VIRTUAL_TIMER *mx_callback_timer;
	size_t num_bytes_available, bytes_read;
	int c, error_flag, debug_level;
	mx_bool_type perform_dns_lookup = TRUE;
	mx_bool_type start_debugger = FALSE;
	mx_bool_type list_pvs = FALSE;
	char *database_name;
	char *acl_name;
	static const char usage[] =
"\n"
"Usage: mxca_server [options ...] 'mx_database_file' 'acl_file'\n"
"\n"
"  Options are:\n"
"    -b              Bypass DNS lookup\n"
"    -d debug_level  Set MX debugging level\n"
"    -D              Start debugger\n"
"    -l              List created EPICS PVs\n"
"\n";

	mx_status_type mx_status;
	error_flag = FALSE;
	debug_level = 0;

	while ((c = getopt(argc, argv, "bd:Dl")) != -1 )
	{
		switch(c) {
		case 'b':	/* Bypass DNS lookup. */
			perform_dns_lookup = FALSE;
			break;
		case 'd':
			debug_level = atoi( optarg );
			break;
		case 'D':
			start_debugger = TRUE;
			break;
		case 'l':
			list_pvs = TRUE;
			break;
		case '?':
			error_flag = TRUE;
			break;
		}
	}

	if ( argc < 3 ) {
		fprintf(stderr, usage);
		exit(1);
	}

	if ( error_flag == TRUE ) {
		fprintf(stderr, usage);
		exit(1);
	}

	database_name = argv[optind];
	acl_name = argv[optind+1];

	mx_set_debug_level(debug_level);

	if ( start_debugger ) {
		mx_start_debugger(NULL);
	}

	/* Register a callback for MX export directives. */

	mx_status = mx_register_export_callback( (char *) "epics",
						mxca_export_callback_shim );

	if ( mx_status.code != MXE_SUCCESS )
		exit(mx_status.code);

	/* Create a channel access server. */

	double scan_period = 1.0;	/* in seconds */
	
	server = new mxcaServer( database_name,
				scan_period, perform_dns_lookup );

	if ( server == NULL ) {
		fprintf(stderr, "Could not create MX CA server.\n" );
		exit(1);
	}

	if ( list_pvs ) {
		/* Print out the names of the created PVs. */

		mxcaPV *PVList_head, *current_PV;

		PVList_head = server->PVList;

		current_PV = PVList_head;

		do {
			mx_info( "PV '%s'", current_PV->getName() );

			current_PV = current_PV->get_next_pv();

		} while ( current_PV != PVList_head );
	}

	/* Initialize the socket connection access control list. */

	if ( strlen( acl_name ) > 0 ) {
		mx_status = mx_setup_connection_acl( server->mx_record_list,
							acl_name );

		if ( mx_status.code != MXE_SUCCESS )
			exit(mx_status.code);
	}

	/*--- Get a pointer to the MX callback pipe. ---*/

	mx_record_list = server->mx_record_list;

	mx_list_head = mx_get_record_list_head_struct( mx_record_list );

	mx_callback_pipe = (MX_PIPE *) mx_list_head->callback_pipe;

	mx_callback_timer = (MX_VIRTUAL_TIMER *) mx_list_head->callback_timer;

	/*--- Start the server scan timer. ---*/

	server->start();

	mx_info( "*** MX channel access server ready. ***" );

	/*--- Loop forever ---*/

	double delay = 0.01;

	while (aitTrue) {
		fileDescriptorManager.process( delay );

		mx_status = mx_pipe_num_bytes_available( mx_callback_pipe,
							&num_bytes_available );

		if ( (mx_status.code == MXE_SUCCESS)
		  && (num_bytes_available > 0) )
		{
			mx_status = mx_pipe_read( mx_callback_pipe,
						(char *) &mx_callback_message,
						sizeof(MX_CALLBACK_MESSAGE *),
						&bytes_read );

			if ( mx_callback_message->callback_type == MXCBT_POLL )
			{
				server->scan( mx_callback_timer );
			}
		}
	}

	return 1;
}

