/*
 * Name:    mxcaPV.cpp
 *
 * Purpose: This file defines the mxcaPV class.
 *
 *-----------------------------------------------------------------------
 *
 * Copyright 2004, 2008-2010 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#define MXCA_PV_ALWAYS_POST_EVENT	FALSE

#define MXCA_PV_TEST_FOR_VALUE_CHANGED	TRUE

#define MXCA_PV_DEBUG			FALSE

#define MXCA_PV_DEBUG_READ		FALSE

#define MXCA_PV_DEBUG_ENUM		FALSE

#define MXCA_PV_DEBUG_INTEREST		FALSE

#define MXCA_PV_DEBUG_EVENT_READY	FALSE

#define MXCA_PV_DEBUG_UNIMPLEMENTED_APPLICATION_TYPES	FALSE

/* FIXME - Currently the unimplemented application types include
 * the graphic, control, and alarm limits, as well as the units.
 */

#include <stdio.h>

#include "mxcaServer.h"

#include "dbMapper.h"
#include "gddApps.h"

extern "C" {
#include "mx_time.h"
#include "mx_hrt.h"
}

static aitEnum
mxca_get_primitive_type_from_mx_field_type( long datatype )
{
	aitEnum primitive_type;

	switch( datatype ) {
	case MXFT_STRING:
		primitive_type = aitEnumString;
		break;
	case MXFT_CHAR:
		primitive_type = aitEnumInt8;
		break;
	case MXFT_UCHAR:
		primitive_type = aitEnumUint8;
		break;
	case MXFT_SHORT:
		primitive_type = aitEnumInt16;
		break;
	case MXFT_USHORT:
		primitive_type = aitEnumUint16;
		break;
	case MXFT_BOOL:
	case MXFT_LONG:
		primitive_type = aitEnumInt32;
		break;
	case MXFT_ULONG:
	case MXFT_HEX:
		primitive_type = aitEnumUint32;
		break;
	case MXFT_FLOAT:
		primitive_type = aitEnumFloat32;
		break;
	case MXFT_DOUBLE:
		primitive_type = aitEnumFloat64;
		break;
	default:
		primitive_type = aitEnumInvalid;
		break;
	}
	return primitive_type;
}

/*-------------*/

mxcaPV::mxcaPV( mxcaServer *epics_server,
		mxcaRecord *epics_record,
		const char *epics_field_name ) : casPV()
{
	this->create( epics_server, epics_record, epics_field_name );
}

mxcaPV::mxcaPV( mxcaRecord *epics_record,
		const char *epics_field_name ) : casPV()
{
	this->create( NULL, epics_record, epics_field_name );
}

mxcaPV::mxcaPV( mxcaServer *epics_server,
		const char *epics_field_name ) : casPV()
{
	this->create( epics_server, NULL, epics_field_name );
}

/*-------------*/

void
mxcaPV::create( mxcaServer *epics_server,
		mxcaRecord *epics_record,
		const char *epics_field_name )
{
	static const char fname[] = "mxcaPV::mxcaPV()";

	mxcaPV *PVList;
	mxcaPV *previousPV, *nextPV;

	if ( epics_field_name == NULL ) {
		(void) mx_error( MXE_NULL_ARGUMENT, fname,
		"The epics_field_name pointer passed was NULL." );

		return;		/* FIXME - replace with throw. */
	}

#if MXCA_PV_DEBUG
	MX_DEBUG(-2,("%s invoked for epics_server = %p, epics_record = %p, "
		"epics_field_name = '%s'",
			fname, epics_server, epics_record, epics_field_name));
#endif
	if ( epics_server == (mxcaServer *) NULL ) {
		if ( epics_record == (mxcaRecord *) NULL ) {
			(void) mx_error( MXE_NULL_ARGUMENT, fname,
			"Both the mxcaServer and the mxcaRecord pointers "
			"are NULL for EPICS field name '%s'.",
				epics_field_name );

			return;		/* FIXME - replace with throw. */
		} else {
			this->epicsServer = epics_record->epicsServer;
		}
	} else {
		this->epicsServer = epics_server;
	}

	this->epicsRecord = epics_record;

	if ( this->epicsRecord == (mxcaRecord *) NULL ) {
		this->mx_record = NULL;
	} else {
		this->mx_record = this->epicsRecord->mx_record;
	}

	/* If the epics_server argument passed to this method was not NULL,
	 * then we take the epics_field_name to be the entire PV name.
	 *
	 * If the epics_server argument passed to this method _WAS_ NULL,
	 * then the epics_field_name argument is taken to be the field name
	 * of a PV belonging to the epics_record object.  We construct that
	 * field name by concatenating the epics_field_name argument to the
	 * record name of the EPICS record.
	 */

	if ( epics_server != (mxcaServer *) NULL ) {
		strlcpy( this->PVname, epics_field_name, sizeof(this->PVname) );
	} else {
		snprintf( this->PVname, sizeof(this->PVname),
			"%s.%s",
			epics_record->epicsRecordName,
			epics_field_name );
	}

#if MXCA_PV_DEBUG
	MX_DEBUG(-2,("%s: EPICS PV name = '%s'\n", fname, this->PVname ));
#endif

	/* Add the new PV to the PV list. */

	PVList = this->epicsServer->PVList;

	if ( PVList == (mxcaPV *) NULL ) {

		/* If there is no PVList yet, point nextPV and previousPV
		 * at the new object.
		 */

		this->nextPV = this;
		this->previousPV = this;
	} else {

		nextPV = PVList;

		previousPV = nextPV->previousPV;

		this->nextPV = nextPV;
		this->previousPV = previousPV;

		nextPV->previousPV = this;
		previousPV->nextPV = this;
	}

	this->mx_record_field = NULL;
	this->epics_value = NULL;
	this->mx_value = NULL;
	this->epics_best_external_type = aitEnumInvalid;
	this->must_copy_data = FALSE;
	this->must_transform_data = FALSE;
	this->value_changed_post_in_progress = FALSE;
	this->monitored = FALSE;
	this->value_changed = FALSE;
	this->value_changed_manual_override = FALSE;
	this->saved_value_changed_test_function = NULL;

#if MXCA_PV_DEBUG
	MX_DEBUG(-2,("%s complete for EPICS PV '%s'\n", fname, this->PVname ));
#endif

	return;
}

mxcaPV::~mxcaPV()
{
#if MXCA_PV_DEBUG
	static const char fname[] = "mxcaPV::~mxcaPV()";

	MX_DEBUG(-2,("%s destructor invoked for PV '%s'", fname, this->PVname));
#endif

	return;
}

/* By default, the Channel Access Server library will destroy any PVs that
 * do not have any channels attached to them.  Given that we precreate all
 * of our PVs, we do not want this to happen.  The standard way of doing
 * this is to provide a 'destroy' method that does nothing.
 */

void
mxcaPV::destroy()
{
#if MXCA_PV_DEBUG
	static const char fname[] = "mxcaPV::destroy()";

	MX_DEBUG(-2,("%s: attempt to destroy PV '%s' intercepted.",
		fname, this->PVname));
#endif

	return;
}

const char *
mxcaPV::getName() const
{
#if MXCA_PV_DEBUG
	static const char fname[] = "mxcaPV::getName()";

	MX_DEBUG(-2,("%s: this->PVname = '%s'", fname, this->PVname));
#endif

	return this->PVname;
}

unsigned
mxcaPV::maxSimultAsyncOps( void ) const
{
#if MXCA_PV_DEBUG
	static const char fname[] = "mxcaPV::maxSimultAsyncOps()";

	MX_DEBUG(-2,("%s invoked for '%s'", fname, this->PVname));
#endif

	return 5000;
}

casChannel *
mxcaPV::createChannel( const casCtx &ctx,
			const char * const user,
			const char * const host )
{
#if MXCA_PV_DEBUG
	static const char fname[] = "mxcaPV::createChannel()";

	MX_DEBUG(-2,("%s invoked for PV '%s', user '%s', host '%s'",
		fname, this->PVname, user, host));
#endif

	mxcaChannel *channel = new mxcaChannel( ctx, user, host, this );

	return channel;
}

void
mxcaPV::genCopy( void *value )
{
	this->epics_value->genCopy( this->epics_best_external_type, value );
}

aitEnum
mxcaPV::bestExternalType() const
{
#if MXCA_PV_DEBUG
	static const char fname[] = "mxcaPV::bestExternalType()";

	MX_DEBUG(-2,("%s: '%s', this->epics_best_external_type = %d",
		fname, this->PVname, this->epics_best_external_type));
#endif

	return this->epics_best_external_type;
}

caStatus
mxcaPV::interestRegister( void )
{
#if MXCA_PV_DEBUG_INTEREST
	static const char fname[] = "mxcaPV::interestRegister()";

	MX_DEBUG(-2,("%s invoked for PV '%s'", fname, this->PVname));
#endif
	this->monitored = TRUE;

	/* If possible, turn on value change tests. */

#if MXCA_PV_TEST_FOR_VALUE_CHANGED
	MX_RECORD_FIELD *field;

	field = this->mx_record_field;

	if ( field != (MX_RECORD_FIELD *) NULL ) {
		this->saved_value_changed_test_function =
				field->value_changed_test_function;

		if ( field->value_changed_test_function == NULL ) {
			field->value_changed_test_function =
				mx_default_test_for_value_changed;
		}
	}
#endif

	return S_casApp_success;
}

void
mxcaPV::interestDelete( void )
{
#if MXCA_PV_DEBUG_INTEREST
	static const char fname[] = "mxcaPV::interestDelete()";

	MX_DEBUG(-2,("%s invoked for PV '%s'", fname, this->PVname));
#endif
	this->monitored = FALSE;

	/* If possible, restore value change tests to their original state. */

#if MXCA_PV_TEST_FOR_VALUE_CHANGED
	MX_RECORD_FIELD *field;

	field = this->mx_record_field;

	if ( field != (MX_RECORD_FIELD *) NULL ) {
		field->value_changed_test_function =
				this->saved_value_changed_test_function;
	}
#endif

	return;
}

void
mxcaPV::event_ready( void )
{
	static const char fname[] = "mxcaPV::event_ready()";

	mx_bool_type value_changed;

#if 0 && MXCA_PV_DEBUG_EVENT_READY
	MX_DEBUG(-2,("%s Invoked for PV '%s' vvvvvvvvvvvvvvvvvvvv",
		fname, this->PVname));
#endif

	if ( this->epics_value == (gdd *) NULL ) {

#if MXCA_PV_DEBUG_EVENT_READY
		MX_DEBUG(-2,("%s: skipping PV '%s' since it has no value.",
			fname, this->PVname));
#endif
		return;
	}

	/* Poll the hardware for its current value. */

	caStatus epics_status = this->read();

	if ( epics_status != S_casApp_success) {
		(void) mx_error( MXE_FUNCTION_FAILED, fname,
		"The attempt to read from PV '%s' failed.  "
		"EPICS status = %d, EPICS error = '%s'.",
			this->PVname, epics_status, ca_message( epics_status ));
	}

	this->set_status_severity_and_timestamp( 0, 0 );

	/* If the value has changed, then post an event. */

#if MXCA_PV_ALWAYS_POST_EVENT
	value_changed = TRUE;
#else
	value_changed = this->value_changed;
#endif
	if ( this->value_changed_manual_override ) {
		value_changed = TRUE;

		this->value_changed_manual_override = FALSE;

#if MXCA_PV_DEBUG_EVENT_READY
		MX_DEBUG(-2,("%s: Manual override requested for PV '%s'",
			fname, this->PVname));
#endif
	}

	if ( value_changed ) {
		this->post_value_changed();
	}

	this->value_changed = FALSE;

#if 0 && MXCA_PV_DEBUG_EVENT_READY
	MX_DEBUG(-2,("%s Complete for PV '%s' ^^^^^^^^^^^^^^^^^^^^^",
		fname, this->PVname));
#endif

	return;
}

mx_bool_type
mxcaPV::post_value_changed( void )
{
#if MXCA_PV_DEBUG_EVENT_READY
	static const char fname[] = "mxcaPV::post_value_changed()";
#endif

	mxcaServer *epics_server;

	if ( this->value_changed_post_in_progress ) {

#if MXCA_PV_DEBUG_EVENT_READY
		MX_DEBUG(-2,("%s skipped for PV '%s'", fname, this->PVname));
#endif
		return FALSE;
	}

#if MXCA_PV_DEBUG_EVENT_READY
	MX_DEBUG(-2,("%s called for PV '%s'", fname, this->PVname));
#endif

	epics_server = this->epicsServer;

	this->postEvent( epics_server->event_mask, *(this->epics_value) );

	return TRUE;
}

void
mxcaPV::set_status_severity_and_timestamp( aitInt16 status, aitInt16 severity )
{
#if MXCA_PV_DEBUG
	static const char fname[] =
		"mxcaPV::set_status_severity_and_timestamp()";
#endif

	struct timespec current_timespec;

#if MXCA_PV_DEBUG
	MX_DEBUG(-2,("%s invoked for PV '%s'", fname, this->PVname));
#endif

	/* If the PV has no 'epics_value' GDD structure, then there is
	 * nothing we can do here.
	 */

	if ( this->epics_value == (gdd *) NULL ) {

#if MXCA_PV_DEBUG
		MX_DEBUG(-2,
		("%s: PV '%s' has no 'epics_value' structure.  Skipping...",
		 	fname, this->PVname));
#endif

		return;
	}

	/* Set EPICS status and severity. */

	this->epics_value->setStatSevr( status, severity );

	/* Set the EPICS timestamp. */

	current_timespec = mx_current_os_time();

	this->epics_value->setTimeStamp( &current_timespec );

#if MXCA_PV_DEBUG
	MX_DEBUG(-2,("%s complete for PV '%s'", fname, this->PVname));
#endif

	return;
}

caStatus
mxcaPV::read( const casCtx &ctx, gdd &gdd_object )
{
#if 1 || MXCA_PV_DEBUG || MXCA_PV_DEBUG_UNIMPLEMENTED_APPLICATION_TYPES
	static const char fname[] = "mxcaPV::read( const casCtx &, gdd & )";
#endif

	caStatus read_status;

#if MXCA_PV_DEBUG
	MX_DEBUG(-2,("%s invoked for PV '%s'", fname, this->PVname));
#endif

	if ( gdd_object.isContainer() ) {

		gdd *current_gdd;
		aitUint32 i;

		aitUint32 num_elements = gdd_object.getDataSizeElements();

		gddContainer &container = (gddContainer &) gdd_object;

		gddCursor cursor = container.getCursor();

		cursor.first();

		for ( i = 0; i < num_elements; i++ ) {
			current_gdd = cursor.current();

#if MXCA_PV_DEBUG
			MX_DEBUG(-2,("%s: examining gdd %lu.",
				fname, (unsigned long) i));
#endif

			/* Recursively invoke the read method on
			 * the current gdd.
			 */

			read_status = this->read( ctx, *current_gdd );

			cursor.next();
		}
		
		return S_casApp_success;
	}

	/* Decide what to do based on the GDD's application type. */

	unsigned application_type = gdd_object.applicationType();

	switch( application_type ) {
	case gddAppType_value:
		/* 'value' is the most commonly used application type. */

		read_status = this->read( gdd_object );
		break;
	case gddAppType_precision:
		read_status = this->read_precision( gdd_object );
		break;
	case gddAppType_enums:
		read_status = this->read_enums( gdd_object );
		break;
	case gddAppType_class:
		(void) mx_error( MXE_NOT_YET_IMPLEMENTED, fname,
		"Support for 'PV Info' is not yet implemented for PV '%s'.",
			this->PVname );

		read_status = S_casApp_noSupport;
		break;
	default:

#if MXCA_PV_DEBUG_UNIMPLEMENTED_APPLICATION_TYPES
		(void) mx_error( MXE_NOT_YET_IMPLEMENTED, fname,
		"Unimplemented application type %u requested for PV '%s'.",
			application_type, this->PVname );
#endif

		read_status = S_casApp_noSupport;
		break;
	}

	return read_status;
}

caStatus
mxcaPV::read( gdd &gdd_object )
{
#if MXCA_PV_DEBUG
	static const char fname[] = "mxcaPV::read( gdd & )";
#endif

	gddApplicationTypeTable& table = gddApplicationTypeTable::AppTable();

#if MXCA_PV_DEBUG
	MX_DEBUG(-2,("%s invoked for PV '%s'", fname, this->PVname));
#endif

	caStatus epics_status = this->read();

	if ( epics_status == S_casApp_success ) {

		/* Copy to the object to send to the client. */

		table.smartCopy( &gdd_object, this->epics_value );

#if MXCA_DEBUG_GDD_DUMP
		gdd_object.dump();
#endif
	}

#if MXCA_PV_DEBUG
	MX_DEBUG(-2,("%s complete for PV '%s'", fname, this->PVname));
#endif

	return epics_status;
}

caStatus
mxcaPV::read( void )
{
#if MXCA_PV_DEBUG
	static const char fname[] = "mxcaPV::read( void )";
#endif

	int need_mx_record_processing;
	mx_status_type mx_status;

#if MXCA_PV_DEBUG
	MX_DEBUG(-2,("%s invoked for PV '%s'", fname, this->PVname));
#endif

	/* Do we need to do MX record processing? */

	if ( ( this->mx_record == (MX_RECORD *) NULL )
	  || ( this->mx_record_field == (MX_RECORD_FIELD *) NULL ) )
	{
		need_mx_record_processing = FALSE;
	} else {
		need_mx_record_processing = TRUE;
	}

	if ( need_mx_record_processing ) {

		/* Read the values from the hardware. */

		mx_status = mx_process_record_field( this->mx_record,
							this->mx_record_field,
							MX_PROCESS_GET,
							&(this->value_changed));

		/* If the MX value must be transformed before sending it
		 * to EPICS, perform the transformation here.
		 */

		if ( this->must_transform_data ) {
			mx_status = this->transform_mx_data( MX_PROCESS_GET );
		}

		/* If this is a non-string gddScalar, we must copy the MX
		 * value to the GDD object.  For strings or gddAtomics,
		 * copying the data is not necessary, since we pointed the
		 * GDD's data pointer to the * MX data with putRef() in the
		 * PV's constructor.
		 */

#if MXCA_PV_DEBUG
		MX_DEBUG(-2,("%s: this->mx_value = %p", fname, this->mx_value));
#endif

		if ( this->must_copy_data ) {
			this->epics_value->genCopy(
				this->epics_best_external_type,
				this->mx_value );
		}
	}

#if MXCA_PV_DEBUG_READ
	MX_DEBUG(-2,("read: PV '%s', proc = %d, changed = %d, copy = %d",
		this->PVname, need_mx_record_processing,
		this->value_changed, this->must_copy_data));
#endif

	/* Make sure the PV status, etc. is up to date. */

	this->set_status_severity_and_timestamp( 0, 0 );

#if MXCA_PV_DEBUG
	MX_DEBUG(-2,("%s complete for PV '%s'", fname, this->PVname));
#endif

	return S_casApp_success;
}

caStatus
mxcaPV::write( const casCtx &ctx, const gdd &gdd_object )
{
#if MXCA_PV_DEBUG
	static const char fname[] =
		"mxcaPV::write( const casCtx &, const gdd & )";

	MX_DEBUG(-2,("%s invoked for PV '%s'", fname, this->PVname));
#endif

	caStatus write_status = this->write( gdd_object );

	return write_status;
}

caStatus
mxcaPV::write( const gdd &gdd_object )
{
#if MXCA_PV_DEBUG
	static const char fname[] = "mxcaPV::write( const gdd & )";
#endif

	gddApplicationTypeTable& table = gddApplicationTypeTable::AppTable();
	mx_status_type mx_status;

#if MXCA_PV_DEBUG
	MX_DEBUG(-2,("%s invoked for PV '%s'", fname, this->PVname));
#endif

#if MXCA_DEBUG_GDD_DUMP
	gdd_object.dump();
#endif

	/* Copy from the object sent by the client. */

	table.smartCopy( this->epics_value, &gdd_object );

	if ( this->mx_value != NULL ) {

		/* Write the values to the hardware. */

		/* If this is a non-string gddScalar, we must copy the GDD
		 * value to the MX value pointer.  For strings or gddAtomics,
		 * copying the data is not necessary, since we pointed the
		 * GDD's data pointer to the * MX data with putRef() in the
		 * PV's constructor.
		 */

#if MXCA_PV_DEBUG
		MX_DEBUG(-2,("%s: this->mx_value = %p", fname, this->mx_value));
#endif

		if ( this->must_copy_data ) {
			this->epics_value->get( this->epics_best_external_type,
						this->mx_value );
		}

		/* If the EPICS value must be transformed before sending it
		 * to MX, perform the transformation here.
		 */

		if ( this->must_transform_data ) {
			mx_status = this->transform_mx_data( MX_PROCESS_PUT );
		}
	}

	caStatus write_status = this->write();

	return write_status;
}

caStatus
mxcaPV::write( void )
{
#if MXCA_PV_DEBUG
	static const char fname[] = "mxcaPV::write( void )";
#endif
	mx_status_type mx_status;

#if MXCA_PV_DEBUG
	MX_DEBUG(-2,("%s invoked for PV '%s'", fname, this->PVname));
#endif

	if ( ( this->mx_record == (MX_RECORD *) NULL )
	  || ( this->mx_record_field == (MX_RECORD_FIELD *) NULL ) )
	{
		/* If the PV is not attached to an MX record field,
		 * then all we do is post the change to EPICS clients.
		 */

		this->post_value_changed();
	} else
	if ( this->mx_value != NULL ) {

		/* Send the value to the MX driver. */

		mx_status = mx_process_record_field( this->mx_record,
						this->mx_record_field,
						MX_PROCESS_PUT,
						&(this->value_changed) );

		if ( this->value_changed ) {
			this->genCopy( this->mx_value );

			this->post_value_changed();
		}
	}

#if MXCA_PV_DEBUG
	MX_DEBUG(-2,("%s complete for PV '%s'", fname, this->PVname));
#endif

	return S_casApp_success;
}

mx_status_type
mxcaPV::transform_mx_data( int direction )
{
#if MXCA_PV_DEBUG
	static const char fname[] = "mxcaPV::transform_mx_data()";
#endif

#if MXCA_PV_DEBUG
	MX_DEBUG(-2,("%s invoked for PV '%s'", fname, this->PVname));
#endif

	/* The default transformation is to make no changes to the data. */

#if MXCA_PV_DEBUG
	MX_DEBUG(-2,("%s complete for PV '%s'", fname, this->PVname));
#endif

	return MX_SUCCESSFUL_RESULT;
}

caStatus
mxcaPV::read_precision( gdd &gdd_object )
{
#if MXCA_PV_DEBUG
	static const char fname[] = "mxcaPV::read_precision()";

	MX_DEBUG(-2,("%s invoked for PV '%s'", fname, this->PVname));
#endif

	/* FIXME: For now we set all precision values to 4. */

	gdd_object.put( 4 );

	return S_casApp_success;
}

caStatus
mxcaPV::read_enums( gdd &gdd_object )
{
	static const char fname[] = "mxcaPV::read_enums()";

	MX_DEBUG(-2,
	("ERROR!!!: %s invoked for PV '%s'.  This should not happen.",
		fname, this->PVname));

	return S_casApp_noSupport;
}

mx_status_type
mxcaPV::configure_internal_pv( aitEnum primitive_type,
				unsigned dimension,
				aitUint32 *size_array,
				const void *initial_value )
{
#if MXCA_PV_DEBUG
	static const char fname[] = "mxcaPV::configure_internal_pv()";
#endif

	int application_type;
	char *string_ptr;

#if MXCA_PV_DEBUG
	MX_DEBUG(-2,("%s invoked for EPICS PV '%s'", fname, this->PVname));
#endif

	application_type = gddAppType_value;

	this->epics_best_external_type = primitive_type;

	if ( dimension == 0 ) {

#if MXCA_PV_DEBUG
		MX_DEBUG(-2,("%s: PV '%s', creating gddScalar( %d, %d )",
			fname, this->PVname, application_type, primitive_type));
#endif

		this->epics_value = new gddScalar( application_type,
							primitive_type );
	} else {

#if MXCA_PV_DEBUG
		MX_DEBUG(-2,
		("%s: PV '%s', creating gddAtomic( %d, %d, %d, %d )",
		 	fname, this->PVname, application_type, primitive_type,
					dimension, size_array[0] ));
#endif

		this->epics_value = new gddAtomic( application_type,
							primitive_type,
							dimension, size_array );
	}

	switch( primitive_type ) {
	case aitEnumString:
		string_ptr = (char *) initial_value;

		this->epics_value->genCopy( primitive_type, &string_ptr );
		break;
	default:
		this->epics_value->genCopy( primitive_type, initial_value );
		break;
	}

#if MXCA_DEBUG_GDD_DUMP
	this->epics_value->dump();
#endif

#if MXCA_PV_DEBUG
	MX_DEBUG(-2,("%s complete for EPICS PV '%s'", fname, this->PVname));
#endif

	return MX_SUCCESSFUL_RESULT;
}

mx_status_type
mxcaPV::configure_mx_field_pv( const char *mx_field_name )
{
	static const char fname[] = "mxcaPV::configure_mx_field_pv()";

	MX_RECORD_FIELD *field, *matching_field;
	MX_RECORD *mx_record;
	long i;
	mx_status_type mx_status;

	if ( mx_field_name == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The mx_field_name pointer passed was NULL." );
	}

	mx_record = this->mx_record;

	if ( mx_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"EPICS PV '%s' does not have an associated MX record.",
			this->PVname );
	}

#if MXCA_PV_DEBUG
	MX_DEBUG(-2,("%s invoked for EPICS PV '%s', MX field '%s.%s'",
		fname, this->PVname, mx_record->name, mx_field_name ));
#endif

	matching_field = NULL;

	for ( i = 0; i < mx_record->num_record_fields; i++ ) {

		field = &(mx_record->record_field_array[i]);

		if ( strcmp( mx_field_name, field->name ) == 0 ) {
			matching_field = field;

			break;
		}
	}

	if ( matching_field == NULL ) {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"Error in EPICS PV '%s'.  The corresponding MX record '%s' "
		"does not have a field named '%s'",
			this->PVname, mx_record->name, mx_field_name );
	}

	mx_status = this->configure_mx_field_pv( matching_field );

	return mx_status;
}

mx_status_type
mxcaPV::configure_mx_field_pv( MX_RECORD_FIELD *mx_record_field )
{
	static const char fname[] = "mxcaPV::configure_mx_field_pv()";

	MX_RECORD *mx_record;
	aitEnum primitive_type;
	unsigned dimension;
	int application_type;
	aitUint32 size_array[1];

	if ( mx_record_field == (MX_RECORD_FIELD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_RECORD_FIELD pointer passed was NULL." );
	}

	mx_record = mx_record_field->record;

	if ( mx_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_RECORD pointer for MX record field %p '%s' is NULL.",
			mx_record_field, mx_record_field->name );
	}

	this->mx_record_field = mx_record_field;

	this->mx_value = mx_get_field_value_pointer( mx_record_field );

	if ( this->mx_record == (MX_RECORD *) NULL ) {
		this->mx_record = mx_record;
	}

	/* Find out what data type this MX field is and create
	 * a matching GDD structure.
	 */

	dimension = mx_record_field->num_dimensions;

	if ( ( dimension != 0 ) && ( dimension != 1 ) ) {
		return mx_error( MXE_UNSUPPORTED, fname,
	"The MX Channel Access server only supports exporting MX record fields "
	"with 0 or 1 dimensions.  The requested field '%s' for record '%s' "
	"has %u dimensions.", mx_record_field->name, mx_record->name,
				dimension );
	}

#if 1
	/* A zero-dimensional string for EPICS is the same thing
	 * as a one-dimensional string for MX.  (I think?)
	 */

	if ( (dimension > 0) && (mx_record_field->datatype == MXFT_STRING) ) {
		dimension = 0;
	}
#endif

	if ( dimension == 1 ) {
		size_array[0] = mx_record_field->dimension[0];
	}

	primitive_type = mxca_get_primitive_type_from_mx_field_type(
				mx_record_field->datatype );

	if ( primitive_type == aitEnumInvalid ) {
		return mx_error( MXE_UNSUPPORTED, fname,
	"MX record field '%s.%s' may not be exported to EPICS since it is "
	"of unsupported type %ld '%s'.",
				mx_record->name, mx_record_field->name,
				mx_record_field->datatype,
			mx_get_field_type_string( mx_record_field->datatype ) );
	}

	this->epics_best_external_type = primitive_type;

	application_type = gddAppType_value;

	/* Now create the gdd object that EPICS uses for the
	 * value of the field.
	 */

	this->must_copy_data = FALSE;

	if ( dimension == 0 ) {

#if MXCA_PV_DEBUG
		MX_DEBUG(-2,("%s: PV '%s', creating gddScalar( %d, %d )",
			fname, this->PVname, application_type, primitive_type));
#endif

		this->epics_value = new gddScalar( application_type,
						primitive_type );

		if ( ( primitive_type != aitEnumString )
		  && ( primitive_type != aitEnumFixedString ) )
		{
			this->must_copy_data = TRUE;
		}
	} else
	if ( dimension == 1 ) {

#if MXCA_PV_DEBUG
		MX_DEBUG(-2,
		("%s: PV '%s', creating gddAtomic( %d, %d, %d, %d )",
			fname, this->PVname, application_type, primitive_type,
			dimension, size_array[0] ));
#endif

		this->epics_value = new gddAtomic( application_type,
						primitive_type,
						dimension, size_array );
	} else {
		return mx_error( MXE_UNSUPPORTED, fname,
			"dimension (%u) is greater than 1.  "
			"We should not be able to get here.", dimension );
	}

	/* If this is an array or string, point the epics_value gdd
	 * object's data pointer to the MX record field data.
	 */

	if ( this->must_copy_data == FALSE ) {
		this->epics_value->putRef( this->mx_value, primitive_type );
	}

#if MXCA_PV_DEBUG
	MX_DEBUG(-2,
("%s: '%s', this->mx_value = %p, epics_data_pointer = %p, primitive_type = %d",
		fname, this->PVname, this->mx_value,
		this->epics_value->dataPointer(), primitive_type));

	MX_DEBUG( 2,("%s completed for EPICS PV '%s', MX field '%s.%s'",
		fname, this->PVname, mx_record->name, mx_field_name ));
#endif

	return MX_SUCCESSFUL_RESULT;
}

void *
mxcaPV::get_mx_value_pointer( void ) const
{
	return this->mx_value;
}

void
mxcaPV::set_mx_value_pointer( void *value_ptr )
{
	this->mx_value = value_ptr;

	if ( value_ptr == NULL ) {
		this->must_copy_data = FALSE;
	} else {
		this->must_copy_data = TRUE;
	}
	return;
}

mxcaPV *
mxcaPV::get_next_pv( void ) const
{
	return this->nextPV;
}

mx_bool_type
mxcaPV::get_value_changed( void )
{
	return this->value_changed;
}

/*--------------------------------------------------------------------------*/

mxcaEnumPV::mxcaEnumPV( mxcaRecord *epics_record, const char *epics_field_name )
			: mxcaPV( epics_record, epics_field_name )
{
	return;
}

mxcaEnumPV::~mxcaEnumPV( void )
{
	return;
}

mx_status_type
mxcaEnumPV::configure_enum_pv( aitEnum16 initial_value,
				unsigned num_enum_values, ... )
{
#if MXCA_PV_DEBUG_ENUM
	static const char fname[] = "mxcaEnumPV::configure_enum_pv()";
#endif

	va_list args;
	unsigned i;
	char *ptr;
	mx_status_type mx_status;

#if MXCA_PV_DEBUG_ENUM
	MX_DEBUG(-2,("%s invoked for EPICS PV '%s'", fname, this->PVname));
#endif

	this->enum_value = initial_value;
	this->num_enums = num_enum_values;

	this->mx_value = &(this->enum_value);

	va_start(args, num_enum_values);

	this->enum_array = new aitFixedString[ this->num_enums ];

	for ( i = 0; i < this->num_enums; i++ ) {
		ptr = va_arg( args, char * );

		strlcpy( this->enum_array[i].fixed_string,
				ptr, sizeof( this->enum_array[i] ) );

#if MXCA_PV_DEBUG_ENUM
		MX_DEBUG(-2,("%s: this->enum_array[%u].fixed_string = '%s'",
			fname, i, this->enum_array[i].fixed_string));
#endif
	}

	va_end(args);

	mx_status = this->configure_internal_pv( aitEnumEnum16, 0,
						NULL, &(this->enum_value) );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	this->epics_value->putConvert( initial_value );

#if 0
	this->epics_value->dump();
#endif

#if MXCA_PV_DEBUG_ENUM
	MX_DEBUG(-2,("%s complete for EPICS PV '%s'", fname, this->PVname));
#endif

	return MX_SUCCESSFUL_RESULT;
}

caStatus
mxcaEnumPV::read_enums( gdd &gdd_object )
{
#if MXCA_PV_DEBUG_ENUM
	static const char fname[] = "mxcaPV::read_enums()";

	MX_DEBUG(-2,("%s invoked for PV '%s'", fname, this->PVname));
#endif

	gdd_object.setDimension( 1 );
	gdd_object.setBound( 0, 0, this->num_enums );
	gdd_object.putRef( this->enum_array );

	return S_casApp_success;
}

caStatus
mxcaEnumPV::write( const gdd &gdd_object )
{
#if MXCA_PV_DEBUG_ENUM
	static const char fname[] = "mxcaEnumPV::write()";
#endif

	aitEnum16 gdd_value;
	caStatus epics_status;

#if MXCA_PV_DEBUG_ENUM
	MX_DEBUG(-2,("%s invoked for PV '%s'", fname, this->PVname));
#endif

	/* Save the value in the this->epics_value structure. */

	epics_status = mxcaPV::write( gdd_object );

	if ( epics_status != S_casApp_success )
		return epics_status;

	/* Get the PV value from the GDD structure so that we may
	 * save a copy of it in this->enum_value.
	 */

	gdd_object.getConvert( gdd_value );

	this->enum_value = gdd_value;

#if MXCA_PV_DEBUG_ENUM
	MX_DEBUG(-2,("%s: PV '%s', this->enum_value = %d",
		fname, this->PVname, this->enum_value));

	MX_DEBUG(-2,("%s complete for PV '%s'", fname, this->PVname));
#endif

	return S_casApp_success;
}

/*--------------------------------------------------------------------------*/

mxcaDummyPV::mxcaDummyPV( mxcaServer *epics_server,
			const char *epics_pvname ) :
			mxcaPV( epics_server, epics_pvname )
{
	;
}

mxcaDummyPV::mxcaDummyPV( mxcaRecord *epics_record,
			const char *epics_pvname ) :
			mxcaPV( epics_record, epics_pvname )
{
	;
}

caStatus
mxcaDummyPV::read( void )
{
	return S_casApp_success;
}

caStatus
mxcaDummyPV::write( const gdd &gdd_object )
{
	return S_casApp_success;
}

