/*
 * Name:    mxcaScalerRecord.cpp
 *
 * Purpose: This file defines the mxcaScalerRecord class.
 *
 *-----------------------------------------------------------------------
 *
 * Copyright 2009 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#define MXCA_SCALER_RECORD_DEBUG	FALSE

#include <iostream>
#include <fstream>

#include "mxcaServer.h"
#include "mxcaScalerRecord.h"

extern "C" {
#include "mx_timer.h"
#include "mx_scaler.h"
#include "mx_analog_input.h"
#include "mx_digital_input.h"
#include "mx_motor.h"
#include "mx_relay.h"
}

mxcaScalerRecord::mxcaScalerRecord( mxcaServer *epics_server,
				const char *epics_record_name,
				MX_RECORD *mx_timer_record ) :
					mxcaRecord( epics_server,
						epics_record_name,
						mx_timer_record )
{
#if MXCA_SCALER_RECORD_DEBUG
	static const char fname[] = "mxcaScalerRecord::mxcaScalerRecord()";
#endif

	mxcaPV *pv;
	mxcaScalerFreqPV *freq_pv;
	mxcaScalerPr1PV *pr1_pv;
	mxcaScalerS1PV *s1_pv;
	mxcaScalerTPV *t_pv;
	mxcaScalerTpPV *tp_pv;
	float float_value;

	const char *channel1_name;

#if MXCA_SCALER_RECORD_DEBUG
	MX_DEBUG(-2,("%s invoked for EPICS scaler record '%s'",
		fname, epics_record_name));
#endif
	this->num_inputs = 0;

	this->cnt_pv = new mxcaScalerCntPV( this );

	pv = new mxcaPV( this, "DTYP" );
	pv->configure_internal_pv( aitEnumString, 0, NULL, "MxScaler" );

	freq_pv = new mxcaScalerFreqPV( this );

	this->g1_pv = new mxcaScalerGnPV( this, "G1" );

	pv = new mxcaPV( this, "PREC" );
	pv->configure_mx_field_pv( "precision" );

	pr1_pv = new mxcaScalerPr1PV( this );

	s1_pv = new mxcaScalerS1PV( this );

	pv = new mxcaPV( this, "EGU" );
	pv->configure_internal_pv( aitEnumString, 0, NULL, "sec" );

	t_pv = new mxcaScalerTPV( this );

	tp_pv = new mxcaScalerTpPV( this );

	pv = new mxcaPV( this, "VERS" );
	float_value = (float) 3.2;
	pv->configure_internal_pv( aitEnumFloat32, 0, NULL, &float_value );

	/*---*/

	if ( strlen(mx_timer_record->label) > 0 ) {
		channel1_name = mx_timer_record->label;
	} else {
		channel1_name = mx_timer_record->name;
	}

	pv = new mxcaPV( this, "NM1" );
	pv->configure_internal_pv( aitEnumString, 0, NULL, channel1_name );

	/* Dummy placeholder PVs. */

	pv = new mxcaDummyPV( this, "DLY" );
	float_value = 0;
	pv->configure_internal_pv( aitEnumFloat32, 0, NULL, &float_value );

	pv = new mxcaDummyPV( this, "RATE" );
	float_value = 0;
	pv->configure_internal_pv( aitEnumFloat32, 0, NULL, &float_value );

#if 0
	char pvname[80];

	snprintf( pvname, sizeof(pvname), "%s_mode.VAL", epics_record_name );

	pv = new mxcaDummyPV( epics_server, pvname );
	long_value = 0;
	pv->configure_internal_pv( aitInt32, 0, NULL, &long_value );
#endif

#if MXCA_SCALER_RECORD_DEBUG
	MX_DEBUG(-2,("%s complete for EPICS scaler record '%s'",
		fname, epics_record_name));
#endif
}

mx_status_type
mxcaScalerRecord::add_input( MX_RECORD *mx_input_record )
{
#if MXCA_SCALER_RECORD_DEBUG
	static const char fname[] = "mxcaScalerRecord::add_input()";
#endif
	mxcaPV *pv;
	mxcaScalerGnPV *gn_pv;
	mxcaScalerPrnPV *prn_pv;
	mxcaScalerSnPV *sn_pv;
	int array_index, channel_number;
	char epics_field_name[40];
	const char *channel_name;

	array_index = this->num_inputs;

	channel_number = this->num_inputs + 2;

#if MXCA_SCALER_RECORD_DEBUG
	MX_DEBUG(-2,("%s: MX input '%s', index = %d, channel = %d.",
		fname, mx_input_record->name, array_index, channel_number ));
#endif

	this->input_array[array_index].mx_input_record = mx_input_record;
	this->input_array[array_index].channel_number = channel_number;

	/*---*/

	snprintf( epics_field_name, sizeof(epics_field_name),
		"G%d", channel_number );

	gn_pv = new mxcaScalerGnPV( this, epics_field_name );

	/*---*/

	if ( strlen(mx_input_record->label) > 0 ) {
		channel_name = mx_input_record->label;
	} else {
		channel_name = mx_input_record->name;
	}

	snprintf( epics_field_name, sizeof(epics_field_name),
		"NM%d", channel_number );

	pv = new mxcaPV( this, epics_field_name );
	pv->configure_internal_pv( aitEnumString, 0, NULL, channel_name );

	/*---*/

	snprintf( epics_field_name, sizeof(epics_field_name),
		"PR%d", channel_number );

	prn_pv = new mxcaScalerPrnPV( this, epics_field_name );

	/*---*/

	snprintf( epics_field_name, sizeof(epics_field_name),
		"S%d", channel_number );

	sn_pv = new mxcaScalerSnPV( this, epics_field_name );

	/*---*/

	this->num_inputs++;

	return MX_SUCCESSFUL_RESULT;
}

void
mxcaScalerRecord::update_number_of_channels( void )
{
	mxcaPV *pv;
	short short_value;

	pv = new mxcaPV( this, "NCH" );
	short_value = this->num_inputs + 1;
	pv->configure_internal_pv( aitEnumInt16, 0, NULL, &short_value );
}

void
mxcaScalerRecord::post_gate_values_changed( void )
{
	mxcaInputInfoStruct *input_struct;
	int i;

	/* Send post events for all of the other gate channels. */

	if ( this->g1_pv != NULL ) {
		this->g1_pv->read();

		this->g1_pv->post_value_changed();
	}

	for ( i = 0; i < this->num_inputs; i++ ) {
		input_struct = &(this->input_array[i]);

		if ( input_struct != NULL ) {
			if ( input_struct->gn_pv != NULL ) {
				input_struct->gn_pv->read();

				input_struct->gn_pv->post_value_changed();
			}
		}
	}

	return;
}

/*-----------------------------------------------------------------------*/

mxcaScalerCntPV::mxcaScalerCntPV( mxcaScalerRecord *epics_scaler_record ) :
			mxcaEnumPV( epics_scaler_record, "CNT" )
{
	enum_value = 0;
	this->configure_enum_pv( enum_value, 2, "Done", "Count" );

	return;
}

caStatus
mxcaScalerCntPV::read( void )
{
	static const char fname[] = "mxcaScalerCntPV::read( void )";

	mxcaScalerRecord *epics_scaler_record;
	mxcaInputInfoStruct *input_struct;
	MX_RECORD *gate_record;
	MX_RECORD_FIELD *busy_field;
	void *busy_value_ptr;
	int gate_channel;
	mx_bool_type value_changed;
	mx_status_type mx_status;

#if MXCA_SCALER_RECORD_DEBUG
	MX_DEBUG(-2,("%s invoked for PV '%s'", fname, this->PVname));
#endif

	epics_scaler_record = (mxcaScalerRecord *) this->epicsRecord;

	if ( epics_scaler_record == (mxcaScalerRecord *) NULL ) {
		(void) mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The epicsRecord pointer for PV '%s' is NULL.",
			this->PVname );

		return S_casApp_undefined;
	}

	gate_channel = epics_scaler_record->gate_channel;

	if ( gate_channel <= 0 ) {
		/* If the gate channel is 0 or less, then explicitly
		 * set the gate channel to 1.
		 */

		epics_scaler_record->gate_channel = gate_channel = 1;
	}

	if ( gate_channel == 1 ) {
		/* Channel 1 is always the timer channel. */

		gate_record = this->mx_record;

		if ( gate_record == (MX_RECORD *) NULL ) {
			(void) mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"The MX timer record pointer for PV '%s' is NULL.",
				this->PVname );

			return S_casApp_undefined;
		}
	} else {
		/* Gate channel numbers above 2 will only be set if they
		 * refer to scaler channels.  The write() method for the
		 * mxcaScalerGnPV class makes sure that this is true.
		 */

		input_struct =
		    &(epics_scaler_record->input_array[gate_channel - 2]);

		gate_record = input_struct->mx_input_record;

		if ( gate_record == (MX_RECORD *) NULL ) {
			(void) mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"The MX scaler record pointer for channel %d "
			"used by PV '%s' is NULL.",
				gate_channel, this->PVname );

			return S_casApp_undefined;
		}
	}

	busy_field = mx_get_record_field( gate_record, "busy" );

	if ( busy_field == (MX_RECORD_FIELD *) NULL ) {
		(void) mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The 'busy' field pointer for PV '%s' is NULL.", this->PVname );

		return S_casApp_undefined;
	}

	mx_status = mx_process_record_field( gate_record, busy_field,
					MX_PROCESS_GET, &value_changed );

	busy_value_ptr = mx_get_field_value_pointer( busy_field );

	this->enum_value = *( (int *) busy_value_ptr );

#if MXCA_SCALER_RECORD_DEBUG
	MX_DEBUG(-2,("%s: enum_value = %d", fname, this->enum_value ));
#endif
	/* 1 and 0 mean the same thing for the MX value as for the
	 * EPICS value, so we can copy the value directly.
	 */

	this->genCopy( this->mx_value );

	/* Make sure the PV status, etc. is up to date. */

	this->set_status_severity_and_timestamp( 0, 0 );

	if ( this->enum_value != this->old_enum_value ) {
		this->value_changed_manual_override = TRUE;
	}

	this->old_enum_value = this->enum_value;

#if MXCA_SCALER_RECORD_DEBUG
	MX_DEBUG(-2,("%s complete for PV '%s'", fname, this->PVname));
#endif

	return S_casApp_success;
}

caStatus
mxcaScalerCntPV::write( const gdd &gdd_object )
{
	static const char fname[] = "mxcaScalerCntPV::write()";

	gddApplicationTypeTable& table = gddApplicationTypeTable::AppTable();
	aitEnum16 gdd_set_value;
	MX_RECORD *mx_timer_record, *mx_input_record;
	int i, gate_channel;
	mxcaScalerRecord *epics_scaler_record;
	mxcaInputInfoStruct *input_struct;

#if MXCA_SCALER_RECORD_DEBUG
	MX_DEBUG(-2,("%s invoked for PV '%s'", fname, this->PVname));
#endif

	/* Copy from the object sent by the client. */

	table.smartCopy( this->epics_value, &gdd_object );

#if MXCA_SCALER_RECORD_DEBUG
	MX_DEBUG(-2,("%s: PV '%s', this->epics_value = %d",
		fname, this->PVname, this->enum_value));
#endif

	/* Get the CNT PV value from the GDD structure so that we may
	 * save a copy of it in this->enum_value.
	 */

	gdd_object.getConvert( gdd_set_value );

	this->enum_value = gdd_set_value;

#if MXCA_SCALER_RECORD_DEBUG
	MX_DEBUG(-2,("%s: PV '%s', this->enum_value = %d",
		fname, this->PVname, this->enum_value));
#endif

	mx_timer_record = this->mx_record;

	if ( mx_timer_record == (MX_RECORD *) NULL ) {
		(void) mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_RECORD pointer for PV '%s' is NULL.",
			this->PVname );

		return S_casApp_undefined;
	}

	epics_scaler_record = (mxcaScalerRecord *) this->epicsRecord;

	if ( epics_scaler_record == (mxcaScalerRecord *) NULL ) {
		(void) mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The epicsRecord pointer for PV '%s' is NULL.",
			this->PVname );

		return S_casApp_undefined;
	}

	/*------------------------------------------------------------------*/
	/* Clear all the input devices and set the 'acquisition_in_progress'
	 * flag for all of them.
	 */

	(void) mx_timer_clear( epics_scaler_record->mx_record );

	for ( i = 0; i < epics_scaler_record->num_inputs; i++ ) {

		input_struct = &(epics_scaler_record->input_array[i]);

		mx_input_record = input_struct->mx_input_record;

		switch( mx_input_record->mx_class ) {
		case MXC_ANALOG_INPUT:
			(void) mx_analog_input_clear( mx_input_record );
			break;
		case MXC_DIGITAL_INPUT:
			(void) mx_digital_input_clear( mx_input_record );
			break;
		case MXC_SCALER:
			(void) mx_scaler_clear( mx_input_record );
			break;
		case MXC_TIMER:
			(void) mx_timer_clear( mx_input_record );
			break;
		case MXC_MOTOR:
		case MXC_RELAY:
			break;
		}

		input_struct->sn_pv->acquisition_in_progress = TRUE;
	}

	/*------------------------------------------------------------------*/
	/* Start the acquisition. */

	gate_channel = epics_scaler_record->gate_channel;

	if ( gate_channel <= 0 ) {
		/* If the gate channel is 0 or less, then currently
		 * there is no gate channel, so we must return 
		 * without doing anything.
		 */
		return S_casApp_success;
	} else
	if ( gate_channel == 1 ) {
		/* Channel 1 is always the timer channel. */

		switch( this->enum_value ) {
		case 0:
			(void) mx_timer_stop( mx_timer_record, NULL );
			break;
		case 1:
			(void) mx_timer_start( mx_timer_record,
					epics_scaler_record->tp_value );
			break;
		default:
			(void) mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"Illegal enum value %d written to PV '%s'.",
				this->enum_value, this->PVname );
			break;
		}
	} else {
		/* Gate channel numbers above 2 will only be set if they
		 * refer to scaler channels.  The write() method for the
		 * mxcaScalerGnPV class makes sure that this is true.
		 */

		input_struct =
		    &(epics_scaler_record->input_array[gate_channel - 2]);

		switch( this->enum_value ) {
		case 0:
			(void) mx_scaler_stop( input_struct->mx_input_record,
						NULL );
			break;
		case 1:
			(void) mx_scaler_start( input_struct->mx_input_record,
						input_struct->preset_count );
			break;
		default:
			(void) mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"Illegal enum value %d written to PV '%s'.",
				this->enum_value, this->PVname );
			break;
		}
	}

	this->post_value_changed();

#if MXCA_SCALER_RECORD_DEBUG
	MX_DEBUG(-2,("%s complete for PV '%s'", fname, this->PVname));
#endif

	return S_casApp_success;
}

/*-----------------------------------------------------------------------*/

mxcaScalerFreqPV::mxcaScalerFreqPV( mxcaScalerRecord *epics_scaler_record ) :
			mxcaPV( epics_scaler_record, "FREQ" )
{
	epics_scaler_record->freq_pv = this;

	this->mx_value = &(epics_scaler_record->freq_value);

	epics_scaler_record->freq_value = 1.0e7;

	this->configure_internal_pv( aitEnumFloat64, 0, NULL, this->mx_value );

	return;
}

caStatus
mxcaScalerFreqPV::read( void )
{
	return S_casApp_success;
}

caStatus
mxcaScalerFreqPV::write( const gdd &gdd_object )
{
	static const char fname[] = "mxcaScalerFreqPV::write()";

	gddApplicationTypeTable& table = gddApplicationTypeTable::AppTable();
	mxcaScalerRecord *epics_scaler_record;
	mxcaScalerTPV *t_pv;
	mxcaScalerTpPV *tp_pv;

	epics_scaler_record = (mxcaScalerRecord *) this->epicsRecord;

	if ( epics_scaler_record == (mxcaScalerRecord *) NULL ) {
		(void) mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The mxcaScalerRecord pointer for PV '%s' is NULL.",
			this->PVname );
	}

	/* Copy from the object sent by the client. */

	table.smartCopy( this->epics_value, &gdd_object );

	this->epics_value->get( this->epics_best_external_type,
				&(epics_scaler_record->freq_value) );

	/* Update the value of T. */

	t_pv = epics_scaler_record->t_pv;

	epics_scaler_record->t_value =
		mx_divide_safely( epics_scaler_record->s1_value,
				epics_scaler_record->freq_value );

	t_pv->genCopy( &(epics_scaler_record->t_value) );

	/* Update the value of TP. */

	tp_pv = epics_scaler_record->tp_pv;

	epics_scaler_record->tp_value =
		mx_divide_safely( epics_scaler_record->pr1_value,
				epics_scaler_record->freq_value );

	tp_pv->genCopy( &(epics_scaler_record->tp_value) );

	this->post_value_changed();

	return S_casApp_success;
}

mx_bool_type
mxcaScalerFreqPV::post_value_changed( void )
{
	mxcaScalerRecord *epics_scaler_record;
	mx_bool_type post_status;

	post_status = mxcaPV::post_value_changed();

	if ( post_status == FALSE )
		return FALSE;

	epics_scaler_record = (mxcaScalerRecord *) this->epicsRecord;

	if ( epics_scaler_record == NULL ) {
		return FALSE;
	}

	this->value_changed_post_in_progress = TRUE;

	if ( epics_scaler_record->t_pv != NULL ) {
		epics_scaler_record->t_pv->post_value_changed();
	}

	if ( epics_scaler_record->tp_pv != NULL ) {
		epics_scaler_record->tp_pv->post_value_changed();
	}

	this->value_changed_post_in_progress = FALSE;

	return TRUE;
}

/*-----------------------------------------------------------------------*/

mxcaScalerPr1PV::mxcaScalerPr1PV( mxcaScalerRecord *epics_scaler_record ) :
			mxcaPV( epics_scaler_record, "PR1" )
{
	epics_scaler_record->pr1_pv = this;

	this->mx_value = &(epics_scaler_record->pr1_value);

	epics_scaler_record->pr1_value = 0;

	this->configure_internal_pv( aitEnumUint32, 0, NULL, this->mx_value );

	return;
}

caStatus
mxcaScalerPr1PV::read( void )
{
	return S_casApp_success;
}

caStatus
mxcaScalerPr1PV::write( const gdd &gdd_object )
{
	static const char fname[] = "mxcaScalerPr1PV::write()";

	gddApplicationTypeTable& table = gddApplicationTypeTable::AppTable();
	mxcaScalerRecord *epics_scaler_record;
	mxcaScalerTpPV *tp_pv;

	epics_scaler_record = (mxcaScalerRecord *) this->epicsRecord;

	if ( epics_scaler_record == (mxcaScalerRecord *) NULL ) {
		(void) mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The mxcaScalerRecord pointer for PV '%s' is NULL.",
			this->PVname );
	}

	/* Copy from the object sent by the client. */

	table.smartCopy( this->epics_value, &gdd_object );

	this->epics_value->get( this->epics_best_external_type,
				&(epics_scaler_record->pr1_value) );

	/* Update the value of TP. */

	tp_pv = epics_scaler_record->tp_pv;

	epics_scaler_record->tp_value =
		mx_divide_safely( epics_scaler_record->pr1_value,
				epics_scaler_record->freq_value );

	tp_pv->genCopy( &(epics_scaler_record->tp_value) );

	this->post_value_changed();

	return S_casApp_success;
}

mx_bool_type
mxcaScalerPr1PV::post_value_changed( void )
{
	mxcaScalerRecord *epics_scaler_record;
	mx_bool_type post_status;

	post_status = mxcaPV::post_value_changed();

	if ( post_status == FALSE )
		return FALSE;

	epics_scaler_record = (mxcaScalerRecord *) this->epicsRecord;

	if ( epics_scaler_record == NULL ) {
		return FALSE;
	}

	this->value_changed_post_in_progress = TRUE;

	if ( epics_scaler_record->tp_pv != NULL ) {
		epics_scaler_record->tp_pv->post_value_changed();
	}

	this->value_changed_post_in_progress = FALSE;

	return TRUE;
}

/*-----------------------------------------------------------------------*/

mxcaScalerS1PV::mxcaScalerS1PV( mxcaScalerRecord *epics_scaler_record ) :
			mxcaPV( epics_scaler_record, "S1" )
{
	epics_scaler_record->s1_pv = this;

	this->mx_value = &(epics_scaler_record->s1_value);

	epics_scaler_record->s1_value = 0;

	this->configure_internal_pv( aitEnumUint32, 0, NULL, this->mx_value );

	this->genCopy( &(epics_scaler_record->s1_value) );

	this->post_value_changed();

	return;
}

caStatus
mxcaScalerS1PV::read( void )
{
	static const char fname[] = "mxcaScalerS1PV::read( void )";

	mxcaScalerRecord *epics_scaler_record;
	mxcaScalerTPV *t_pv;
	MX_RECORD *mx_timer_record;
	MX_RECORD_FIELD *mx_timer_value_field;
	MX_TIMER *mx_timer;
	double elapsed_time;
	mx_bool_type value_changed;
	mx_status_type mx_status;

	epics_scaler_record = (mxcaScalerRecord *) this->epicsRecord;

	if ( epics_scaler_record == (mxcaScalerRecord *) NULL ) {
		(void) mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The mxcaScalerRecord pointer for PV '%s' is NULL.",
			this->PVname );
	}

	mx_timer_record = this->mx_record;

	if ( mx_timer_record == (MX_RECORD *) NULL ) {
		(void) mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_RECORD pointer for PV '%s' is NULL.", this->PVname );

		return S_casApp_undefined;
	}

	mx_timer_value_field = mx_get_record_field( mx_timer_record, "value" );

	if ( mx_timer_value_field == (MX_RECORD_FIELD *) NULL ) {
		(void) mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The timer 'value' field pointer for PV '%s' is NULL.",
			this->PVname );

		return S_casApp_undefined;
	}

	mx_timer = (MX_TIMER *) mx_timer_record->record_class_struct;

	if ( mx_timer == (MX_TIMER *) NULL ) {
		(void) mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_TIMER pointer for PV '%s' is NULL.", this->PVname );

		return S_casApp_undefined;
	}

#if 1
	/* FIXME - The value_changed_test_function should already have
	 * been set to a non-NULL value before we get here.
	 */

	if ( mx_timer_value_field->value_changed_test_function == NULL ) {
		mx_timer_value_field->value_changed_test_function
					= mx_default_test_for_value_changed;
	}

#endif

	/* Process the record. */

	mx_status = mx_process_record_field( mx_timer_record,
						mx_timer_value_field,
						MX_PROCESS_GET,
						&value_changed );

	elapsed_time = mx_timer->last_measurement_time - mx_timer->value;

	epics_scaler_record->t_value = elapsed_time;

	epics_scaler_record->s1_value =
		mx_round( elapsed_time * epics_scaler_record->freq_value );

#if MXCA_SCALER_RECORD_DEBUG
	MX_DEBUG(-2,("%s: t_value = %g, freq_value = %g, s1_value = %lu",
		fname, epics_scaler_record->t_value,
		epics_scaler_record->freq_value,
		epics_scaler_record->s1_value));
#endif

	this->genCopy( &(epics_scaler_record->s1_value) );

	t_pv = epics_scaler_record->t_pv;

	t_pv->genCopy( &(epics_scaler_record->t_value) );

	/* Make sure the PV status, etc. is up to date. */

	this->set_status_severity_and_timestamp( 0, 0 );

	if ( value_changed ) {
		this->post_value_changed();
	}

	return S_casApp_success;
}

caStatus
mxcaScalerS1PV::write( const gdd &gdd_object )
{
	return S_casApp_success;
}

mx_bool_type
mxcaScalerS1PV::post_value_changed( void )
{
	mxcaScalerRecord *epics_scaler_record;
	mx_bool_type post_status;

	post_status = mxcaPV::post_value_changed();

	if ( post_status == FALSE )
		return FALSE;

	epics_scaler_record = (mxcaScalerRecord *) this->epicsRecord;

	if ( epics_scaler_record == NULL ) {
		return FALSE;
	}

	this->value_changed_post_in_progress = TRUE;

	if ( epics_scaler_record->t_pv != NULL ) {
		epics_scaler_record->t_pv->post_value_changed();
	}

	this->value_changed_post_in_progress = FALSE;

	return TRUE;
}

/*-----------------------------------------------------------------------*/

mxcaScalerTPV::mxcaScalerTPV( mxcaScalerRecord *epics_scaler_record ) :
			mxcaPV( epics_scaler_record, "T" )
{
	epics_scaler_record->t_pv = this;

	this->mx_value = &(epics_scaler_record->t_value);

	epics_scaler_record->t_value = 0;

	this->configure_internal_pv( aitEnumFloat64, 0, NULL, this->mx_value );

	this->genCopy( &(epics_scaler_record->t_value) );

	this->post_value_changed();

	return;
}

caStatus
mxcaScalerTPV::read( void )
{
	return S_casApp_success;
}

caStatus
mxcaScalerTPV::write( const gdd &gdd_object )
{
	return S_casApp_success;
}

/*-----------------------------------------------------------------------*/

mxcaScalerTpPV::mxcaScalerTpPV( mxcaScalerRecord *epics_scaler_record ) :
			mxcaPV( epics_scaler_record, "TP" )
{
	epics_scaler_record->tp_pv = this;

	this->mx_value = &(epics_scaler_record->tp_value);

	epics_scaler_record->tp_value = 0;

	this->configure_internal_pv( aitEnumFloat64, 0, NULL, this->mx_value );

	return;
}

caStatus
mxcaScalerTpPV::read( void )
{
	return S_casApp_success;
}

caStatus
mxcaScalerTpPV::write( const gdd &gdd_object )
{
	static const char fname[] = "mxcaScalerTpPV::write()";

	gddApplicationTypeTable& table = gddApplicationTypeTable::AppTable();
	mxcaScalerRecord *epics_scaler_record;
	mxcaScalerPr1PV *pr1_pv;

	epics_scaler_record = (mxcaScalerRecord *) this->epicsRecord;

	if ( epics_scaler_record == (mxcaScalerRecord *) NULL ) {
		(void) mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The mxcaScalerRecord pointer for PV '%s' is NULL.",
			this->PVname );
	}

	/* Copy from the object sent by the client. */

	table.smartCopy( this->epics_value, &gdd_object );

	this->epics_value->get( this->epics_best_external_type,
				&(epics_scaler_record->tp_value) );

	/* Update the value of PR1. */

	pr1_pv = epics_scaler_record->pr1_pv;

	epics_scaler_record->pr1_value = mx_round(
	    epics_scaler_record->tp_value * epics_scaler_record->freq_value );

	pr1_pv->genCopy( &(epics_scaler_record->pr1_value) );

	this->post_value_changed();

	return S_casApp_success;
}

mx_bool_type
mxcaScalerTpPV::post_value_changed( void )
{
	mxcaScalerRecord *epics_scaler_record;
	mx_bool_type post_status;

	post_status = mxcaPV::post_value_changed();

	if ( post_status == FALSE )
		return FALSE;

	epics_scaler_record = (mxcaScalerRecord *) this->epicsRecord;

	if ( epics_scaler_record == NULL ) {
		return FALSE;
	}

	this->value_changed_post_in_progress = TRUE;

	if ( epics_scaler_record->pr1_pv != NULL ) {
		epics_scaler_record->pr1_pv->post_value_changed();
	}

	this->value_changed_post_in_progress = FALSE;

	return TRUE;
}

/*=======================================================================*/

mxcaScalerGnPV::mxcaScalerGnPV( mxcaScalerRecord *epics_scaler_record,
				const char *epics_field_name ) :
			mxcaEnumPV( epics_scaler_record, epics_field_name )
{
	static const char fname[] = "mxcaScalerGnPV::mxcaScalerGnPV()";

	mxcaInputInfoStruct *input_struct;
	int i, num_items;

	this->mx_value = &(this->enum_value);

	num_items = sscanf( epics_field_name, "G%d", &(this->channel_number) );

	if ( num_items != 1 ) {
		(void) mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"The supplied epics field name '%s' does not match the "
		"required format of 'Gn' where 'n' is an integer from "
		"1 to the maximum channel number of the scaler.",
			epics_field_name );

		/* FIXME: Throw an exception here. */

		return;
	}

	if ( this->channel_number == 1 ) {
		this->enum_value = 1;

		epics_scaler_record->g1_pv = this;
	} else {
		this->enum_value = 0;

		i = this->channel_number - 2;

		input_struct = &(epics_scaler_record->input_array[i]);

		if ( input_struct != NULL ) {
			input_struct->gn_pv = this;
		}
	}

	this->configure_enum_pv( this->enum_value, 2, "N", "Y" );

	return;
}

caStatus
mxcaScalerGnPV::read( void )
{
	static const char fname[] = "mxcaScalerGnPV::read( void )";

	mxcaScalerRecord *epics_scaler_record;

#if MXCA_SCALER_RECORD_DEBUG
	MX_DEBUG(-2,("%s invoked for PV '%s'", fname, this->PVname));
#endif

	epics_scaler_record = (mxcaScalerRecord *) this->epicsRecord;

	if ( epics_scaler_record == (mxcaScalerRecord *) NULL ) {
		(void) mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The epicsRecord pointer for PV '%s' is NULL.",
			this->PVname );

		return S_casApp_undefined;
	}

	if ( epics_scaler_record->gate_channel == this->channel_number ) {
		this->enum_value = 1;
	} else {
		this->enum_value = 0;
	}

#if MXCA_SCALER_RECORD_DEBUG
	MX_DEBUG(-2,
	("%s: PV '%s' gate_channel = %d, channel_number = %d, enum_value = %d",
		fname, this->PVname, 
		epics_scaler_record->gate_channel,
		this->channel_number, this->enum_value));
#endif

	this->genCopy( this->mx_value );

	/* Make sure the PV status, etc. is up to date. */

	this->set_status_severity_and_timestamp( 0, 0 );

#if MXCA_SCALER_RECORD_DEBUG
	MX_DEBUG(-2,("%s complete for PV '%s'", fname, this->PVname));
#endif

	return S_casApp_success;
}

caStatus
mxcaScalerGnPV::write( const gdd &gdd_object )
{
	static const char fname[] = "mxcaScalerGnPV::write()";

	gddApplicationTypeTable& table = gddApplicationTypeTable::AppTable();
	aitEnum16 gdd_set_value;
	mxcaScalerRecord *epics_scaler_record;
	mxcaInputInfoStruct *input_struct;

#if MXCA_SCALER_RECORD_DEBUG
	MX_DEBUG(-2,("%s invoked for PV '%s'", fname, this->PVname));
#endif

	/* Copy from the object sent by the client. */

	table.smartCopy( this->epics_value, &gdd_object );

#if MXCA_SCALER_RECORD_DEBUG
	MX_DEBUG(-2,("%s: PV '%s', this->epics_value = %d",
		fname, this->PVname, this->enum_value));
#endif

	/* Get the Gn PV value from the GDD structure so that we may
	 * save a copy of it in this->enum_value.
	 */

	gdd_object.getConvert( gdd_set_value );

	this->enum_value = gdd_set_value;

#if MXCA_SCALER_RECORD_DEBUG
	MX_DEBUG(-2,("%s: PV '%s', this->enum_value = %d",
		fname, this->PVname, this->enum_value));
#endif

	epics_scaler_record = (mxcaScalerRecord *) this->epicsRecord;

	if ( epics_scaler_record == (mxcaScalerRecord *) NULL ) {
		(void) mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The epicsRecord pointer for PV '%s' is NULL.",
			this->PVname );

		return S_casApp_undefined;
	}


	if ( this->enum_value != 1 ) {
		epics_scaler_record->gate_channel = -1;
	} else {
		if ( this->channel_number == 1 ) {
			/* Channel 1 is always a timer, so we can 
			 * unconditionally set the gate channel to it.
			 */

			epics_scaler_record->gate_channel = 1;
		} else {
			/* If this is not channel 1, then make sure that
			 * only scaler channels can be used as a gate.
			 */

			input_struct =
		    &(epics_scaler_record->input_array[this->channel_number-2]);

			if ( input_struct->mx_input_record->mx_class
				== MXC_SCALER )
			{
				epics_scaler_record->gate_channel =
					this->channel_number;
			} else {
				epics_scaler_record->gate_channel = -1;
			}
		}
	}

	epics_scaler_record->post_gate_values_changed();

#if MXCA_SCALER_RECORD_DEBUG
	MX_DEBUG(-2,("%s complete for PV '%s'", fname, this->PVname));
#endif

	return S_casApp_success;
}

/*-----------------------------------------------------------------------*/

mxcaScalerPrnPV::mxcaScalerPrnPV( mxcaScalerRecord *epics_scaler_record,
					const char *epics_field_name ) :
				mxcaPV( epics_scaler_record, epics_field_name )
{
	static const char fname[] = "mxcaScalerPrnPV::mxcaScalerPrnPV()";

	mxcaInputInfoStruct *input_struct;
	int num_items;

	num_items = sscanf( epics_field_name, "PR%d", &(this->channel_number) );

	if ( num_items != 1 ) {
		(void) mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"The supplied epics field name '%s' does not match the "
		"required format of 'PRn' where 'n' is an integer from "
		"1 to the maximum channel number of the scaler.",
			epics_field_name );

		/* FIXME: Throw an exception here. */

		return;
	}

	input_struct =
		&(epics_scaler_record->input_array[this->channel_number-2]);

	this->mx_value = &(input_struct->preset_count);

	input_struct->preset_count = 0;

	this->configure_internal_pv( aitEnumUint32, 0, NULL, this->mx_value );

	return;
}

caStatus
mxcaScalerPrnPV::read( void )
{
	return S_casApp_success;
}

caStatus
mxcaScalerPrnPV::write( const gdd &gdd_object )
{
	static const char fname[] = "mxcaScalerPrnPV::write()";

	gddApplicationTypeTable& table = gddApplicationTypeTable::AppTable();
	mxcaScalerRecord *epics_scaler_record;
	mxcaInputInfoStruct *input_struct;

	epics_scaler_record = (mxcaScalerRecord *) this->epicsRecord;

	if ( epics_scaler_record == (mxcaScalerRecord *) NULL ) {
		(void) mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The mxcaScalerRecord pointer for PV '%s' is NULL.",
			this->PVname );
	}

	input_struct =
		&(epics_scaler_record->input_array[this->channel_number-2]);

	/* Copy from the object sent by the client. */

	table.smartCopy( this->epics_value, &gdd_object );

	this->epics_value->get( this->epics_best_external_type,
				&(input_struct->preset_count) );

	this->post_value_changed();

	return S_casApp_success;
}

/*-----------------------------------------------------------------------*/

mxcaScalerSnPV::mxcaScalerSnPV( mxcaScalerRecord *epics_scaler_record,
					const char *epics_field_name ) :
				mxcaPV( epics_scaler_record, epics_field_name )
{
	static const char fname[] = "mxcaScalerSnPV::mxcaScalerSnPV()";

	mxcaInputInfoStruct *input_struct;
	int num_items;
	MX_RECORD *mx_input_record;
	MX_RECORD_FIELD *mx_value_field;

	num_items = sscanf( epics_field_name, "S%d", &(this->channel_number) );

	if ( num_items != 1 ) {
		(void) mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"The supplied epics field name '%s' does not match the "
		"required format of 'Sn' where 'n' is an integer from "
		"1 to the maximum channel number of the scaler.",
			epics_field_name );

		/* FIXME: Throw an exception here. */

		return;
	}

	input_struct =
		&(epics_scaler_record->input_array[this->channel_number-2]);

	input_struct->sn_pv = this;

	this->mx_value = &(input_struct->value);

	input_struct->value = 0;

	this->configure_internal_pv( aitEnumUint32, 0, NULL, this->mx_value );

	/*---*/

	this->cnt_pv = (mxcaScalerCntPV *) this->epicsRecord->findPV( "CNT" );

	if ( this->cnt_pv == NULL ) {
		(void) mx_error( MXE_NOT_FOUND, fname,
		"Did not find a CNT PV for EPICS scaler record '%s'.",
			this->epicsRecord->get_epics_record_name() );

		/* FIXME: Throw an exception here. */

		return;
	}

	/*---*/

	mx_input_record = input_struct->mx_input_record;

	switch( mx_input_record->mx_class ) {
	case MXC_ANALOG_INPUT:
		mx_value_field = mx_get_record_field(mx_input_record, "value");
		break;
	case MXC_DIGITAL_INPUT:
		mx_value_field = mx_get_record_field(mx_input_record, "value");
		break;
	case MXC_MOTOR:
		mx_value_field =
			mx_get_record_field(mx_input_record, "position");
		break;
	case MXC_RELAY:
		mx_value_field =
			mx_get_record_field(mx_input_record, "relay_status");
		break;
	case MXC_SCALER:
		mx_value_field = mx_get_record_field(mx_input_record, "value");
		break;
	case MXC_TIMER:
		mx_value_field = mx_get_record_field(mx_input_record, "value");
		break;
	default:
		(void) mx_error( MXE_UNSUPPORTED, fname,
		"MX record '%s' is not supported by EPICS scaler record '%s' "
		"since it is of unsupported type '%s'.",
			mx_input_record->name,
			epics_scaler_record->get_epics_record_name(),
			mx_get_driver_name( mx_input_record ) );
		break;
	}

	input_struct->mx_value_field = mx_value_field;

	return;
}

caStatus
mxcaScalerSnPV::read( void )
{
	static const char fname[] = "mxcaScalerSnPV::read( void )";

	mxcaScalerRecord *epics_scaler_record;
	mxcaInputInfoStruct *input_struct;
	MX_RECORD *mx_input_record;
	MX_ANALOG_INPUT *ainput;
	MX_DIGITAL_INPUT *dinput;
	MX_MOTOR *motor;
	MX_RELAY *relay;
	MX_SCALER *scaler;
	MX_TIMER *timer;
	mx_bool_type value_changed;
	mx_status_type mx_status;

	/* Only read from the scalers while an acquisition is in progress.
	 * The CNT PV is responsible for setting the 'acquisition_in_progress'
	 * flag at the start of a measurement.
	 */

#if MXCA_SCALER_RECORD_DEBUG
	MX_DEBUG( 2,("%s: PV '%s', acquisition_in_progress = %d",
		fname, this->PVname, this->acquisition_in_progress));
#endif

	if ( this->acquisition_in_progress == FALSE ) {
		return S_casApp_success;
	}

	/*------*/

	epics_scaler_record = (mxcaScalerRecord *) this->epicsRecord;

	if ( epics_scaler_record == (mxcaScalerRecord *) NULL ) {
		(void) mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The mxcaScalerRecord pointer for PV '%s' is NULL.",
			this->PVname );
	}

	input_struct =
		&(epics_scaler_record->input_array[this->channel_number-2]);

	mx_input_record = input_struct->mx_input_record;

	if ( mx_input_record == (MX_RECORD *) NULL ) {
		(void) mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_RECORD pointer for PV '%s' is NULL.", this->PVname );

		return S_casApp_undefined;
	}

	mx_status = mx_process_record_field( mx_input_record,
						input_struct->mx_value_field,
						MX_PROCESS_GET,
						&value_changed );

	if ( mx_status.code != MXE_SUCCESS ) {
		/* FIXME - Throw an exception here. */

		return S_casApp_undefined;
	}

	switch( mx_input_record->mx_class ) {
	case MXC_ANALOG_INPUT:
		ainput =
		    (MX_ANALOG_INPUT *) mx_input_record->record_class_struct;

		input_struct->value =
		    mx_round( ainput->value * epics_scaler_record->freq_value );
		break;
	case MXC_DIGITAL_INPUT:
		dinput =
		    (MX_DIGITAL_INPUT *) mx_input_record->record_class_struct;

		input_struct->value = dinput->value;
		break;
	case MXC_MOTOR:
		motor = (MX_MOTOR *) mx_input_record->record_class_struct;

		input_struct->value =
		  mx_round( motor->position * epics_scaler_record->freq_value );
		break;
	case MXC_RELAY:
		relay = (MX_RELAY *) mx_input_record->record_class_struct;

		input_struct->value = relay->relay_status;
		break;
	case MXC_SCALER:
		scaler = (MX_SCALER *) mx_input_record->record_class_struct;

		input_struct->value = scaler->value;
		break;
	case MXC_TIMER:
		timer = (MX_TIMER *) mx_input_record->record_class_struct;

		input_struct->value =
		  mx_round( timer->value * epics_scaler_record->freq_value );
		break;
	default:
		(void) mx_error( MXE_UNSUPPORTED, fname,
		"MX record '%s' is not supported by EPICS scaler record '%s' "
		"since it is of unsupported type '%s'.",
			mx_input_record->name,
			epics_scaler_record->get_epics_record_name(),
			mx_get_driver_name( mx_input_record ) );
		break;
	}

#if MXCA_SCALER_RECORD_DEBUG
	MX_DEBUG(-2,("%s: PV '%s', value = %lu",
		fname, this->PVname, input_struct->value));
#endif

	this->genCopy( &(input_struct->value) );

	/* Make sure the PV status, etc. is up to date. */

	this->set_status_severity_and_timestamp( 0, 0 );

	this->post_value_changed();

	/* Clear the 'acquisition_in_progress' flag if the measurement
	 * cycle is complete.
	 */

	aitEnum16 cnt_value;

	cnt_value = this->cnt_pv->enum_value;

#if MXCA_SCALER_RECORD_DEBUG
	MX_DEBUG(-2,("%s: cnt_value = %d", fname, cnt_value));
#endif

	if ( cnt_value == 0 ) {
		this->acquisition_in_progress = FALSE;
	}

	return S_casApp_success;
}

caStatus
mxcaScalerSnPV::write( const gdd &gdd_object )
{
	return S_casApp_success;
}

