/*
 * Name:    mxcaServer.h
 *
 * Purpose: This header file defines the mxcaServer class.
 *
 *-----------------------------------------------------------------------
 *
 * Copyright 2004, 2008-2009 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#define MXCA_DEBUG_GDD_DUMP		FALSE

/* Prevent the inclusion of troublesome include files. */

#if defined(OS_WIN32)

   /* Do not include <osdTime.h> */
#  define osdTimeh

   /* Use <winsock2.h> instead of <winsock.h> */
#  include <winsock2.h>
#  define _WINSOCKAPI_

#endif /* OS_WIN32 */

#include <epicsVersion.h>
#include <aitTypes.h>
#include <casdef.h>
#include <gddApps.h>
#include <gddAppFuncTable.h>

extern "C" {
#include "mx_util.h"
#include "mx_record.h"
#include "mx_driver.h"
#include "mx_hash_table.h"
#include "mx_socket.h"
#include "mx_process.h"
#include "mx_security.h"
#include "mx_export.h"
#include "mx_pipe.h"
#include "mx_callback.h"

MX_API mx_status_type mxca_export_callback_shim( MX_RECORD *, char * );
}

class mxcaServer;	// Forward declaration
class mxcaPV;		// Forward declaration

class mxcaRecord {
    friend class mxcaServer;
    friend class mxcaPV;

    public:
    	mxcaRecord ( mxcaServer *epics_server,
			const char *epics_record_name,
			MX_RECORD *mx_record );

	mx_status_type setupRecordPVs( void );

	const char *get_epics_record_name( void );

	mxcaServer *epicsServer;
	MX_RECORD *mx_record;

	mxcaPV *findPV( const char *field_name );

    private:
	char epicsRecordName[80];
};

class mxcaPV : public casPV {
    friend class mxcaServer;
    friend class mxcaChannel;

    public:
	mxcaPV( mxcaServer *epics_server,
		mxcaRecord *epics_record,
		const char *epics_field_name );

	mxcaPV( mxcaRecord *epics_record,
		const char *epics_field_name );

	mxcaPV( mxcaServer *epics_server,
		const char *epics_pv_name );

	~mxcaPV( void );

	void destroy( void );

	virtual const char *getName( void ) const;

	virtual unsigned maxSimultAsyncOps( void ) const;

	virtual casChannel *createChannel( const casCtx &ctx,
					const char * const user,
					const char * const host );

	virtual caStatus read( const casCtx &ctx, gdd &gdd_value );

	virtual caStatus read( gdd &gdd_object );

	virtual caStatus read( void );

	virtual caStatus write( const casCtx &ctx, const gdd &gdd_value );

	virtual caStatus write( const gdd &gdd_object );

	virtual caStatus write( void );

	virtual caStatus read_precision( gdd &gdd_object );

	virtual caStatus read_enums( gdd &gdd_object );

	virtual void genCopy( void *value );

	virtual aitEnum bestExternalType( void ) const;

	virtual caStatus interestRegister( void );

	virtual void interestDelete( void );

	virtual void event_ready( void );

	virtual mx_bool_type post_value_changed( void );

	virtual void set_status_severity_and_timestamp( aitInt16 status,
							aitInt16 severity );

	virtual mx_status_type transform_mx_data( int direction );

	virtual mx_status_type configure_internal_pv( aitEnum datatype,
						unsigned num_dimensions,
						aitUint32 *size_array,
						const void *initial_value );

	virtual mx_status_type configure_mx_field_pv(
						const char *mx_field_name );

	virtual mx_status_type configure_mx_field_pv(
					MX_RECORD_FIELD *mx_record_field );

	virtual void *get_mx_value_pointer( void ) const;

	virtual void set_mx_value_pointer( void * );

	virtual mxcaPV *get_next_pv( void ) const;

	virtual mx_bool_type get_value_changed( void );

    protected:
	char PVname[80];

	mx_bool_type monitored;
	mx_bool_type value_changed;

	mx_bool_type value_changed_manual_override;

	mxcaServer *epicsServer;
	mxcaRecord *epicsRecord;

	MX_RECORD *mx_record;
	MX_RECORD_FIELD *mx_record_field;

	gdd *epics_value;
	void *mx_value;

	aitEnum epics_best_external_type;

	mx_bool_type must_transform_data;
	mx_bool_type must_copy_data;

	mx_bool_type value_changed_post_in_progress;

    private:
	mxcaPV *previousPV;
	mxcaPV *nextPV;

	mx_status_type (*saved_value_changed_test_function)(
					MX_RECORD_FIELD *, mx_bool_type * );

	void create( mxcaServer *epics_server,
		mxcaRecord *epics_record,
		const char *epics_field_name );

};

class mxcaEnumPV : public mxcaPV {
    public:
	mxcaEnumPV( mxcaRecord *epics_record,
			const char *epics_field_name );

	~mxcaEnumPV( void );

	virtual mx_status_type configure_enum_pv( aitEnum16 initial_value,
						unsigned num_enum_values, ... );

	virtual caStatus read_enums( gdd &gdd_object );

	virtual caStatus write( const gdd &gdd_object );

	aitEnum16 enum_value;

	aitEnum16 old_enum_value;

    protected:
	unsigned num_enums;
	aitFixedString *enum_array;
};

class mxcaDummyPV : public mxcaPV {
    public:
	mxcaDummyPV( mxcaServer *epics_server,
			const char *epics_field_name );

	mxcaDummyPV( mxcaRecord *epics_record,
			const char *epics_field_name );

	caStatus read( void );
	caStatus write( const gdd &gdd_object );
};

class mxcaChannel : public casChannel {
    public:
	mxcaChannel( const casCtx &ctx,
			const char * const user,
			const char * const host,
			mxcaPV *channel_pv );
	~mxcaChannel( void );

	virtual bool readAccess( void ) const;
	virtual bool writeAccess( void ) const;
	virtual void setOwner( const char * const user,
				const char * const host );

    private:
	mxcaPV *pv;
	const char *username;
	const char *hostname;
	mx_bool_type read_access_permitted;
	mx_bool_type write_access_permitted;
};

class mxcaServer : public caServer {
    friend class mxcaChannel;

    public:
	mxcaServer( const char *mx_database_filename,
			double scan_period,
			mx_bool_type perform_dns_lookup );

	pvExistReturn pvExistTest( const casCtx &ctx, const char *pPVName );

	pvAttachReturn pvAttach( const casCtx &ctx, const char *pPVName );

	MX_RECORD *mx_record_list;
	mxcaPV *PVList;

	mxcaRecord *findRecord( const char *pRecordName );

	mxcaPV *findPV( const char *pPVName );

	void start( void );

	void scan( MX_VIRTUAL_TIMER *callback_timer );

	casEventMask event_mask;

	mx_bool_type perform_dns_lookup;

    private:
	mx_status_type initializePVList( void );

	mxcaRecord *epicsRecordList;

	MX_HASH_TABLE *hash_table;

	MX_CALLBACK_MESSAGE *poll_callback_message;

	double scan_period;
};

