/*
 * Name:    mxca_export.cpp
 *
 * Purpose: MX !export callback handler.
 *
 *-----------------------------------------------------------------------
 *
 * Copyright 2009 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#define MXCA_EXPORT_DEBUG	FALSE

#include <iostream>
#include <fstream>

#define __MX_LIBRARY__

#include "mxcaServer.h"
#include "mxcaMotorRecord.h"
#include "mxcaScalerRecord.h"

/*----------------------------------------------------------------------*/

static mxcaServer *
mxca_get_epics_server( MX_RECORD *record_list )
{
	static const char fname[] = "mxca_get_epics_server()";

	MX_LIST_HEAD *list_head_struct;
	mxcaServer *epics_server;

	if ( record_list == (MX_RECORD *) NULL ) {
		(void) mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_RECORD pointer passed was NULL." );

		return NULL;
	}

	list_head_struct = mx_get_record_list_head_struct( record_list );

	if ( list_head_struct == (MX_LIST_HEAD *) NULL ) {
		(void) mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_LIST_HEAD pointer for record list %p is NULL.",
			record_list );
	}

	epics_server = (mxcaServer *) list_head_struct->application_ptr;

	return epics_server;
}

/*----------------------------------------------------------------------*/

static mx_status_type
mxca_export_epics_pv( mxcaServer *epics_server,
			MX_RECORD *record_list_head,
			char *buffer, int argc, char **argv )
{
	static const char fname[] = "mxca_export_epics_pv()";

	mxcaRecord *epics_record;
	mxcaPV *epics_pv;
	MX_RECORD *mx_record;
	MX_RECORD_FIELD *mx_field;
	char *record_name, *field_name, *dot_ptr;
	char *pv_name, *rf_name;
	char value_name[] = "value";

#if MXCA_EXPORT_DEBUG
	MX_DEBUG(-2,("%s invoked for '%s'", fname, buffer));
#endif

	if ( argc != 5 ) {
		return mx_error( MXE_UNPARSEABLE_STRING, fname,
		"Illegal number of arguments (%d) for 'epics_motor' export "
		"statement '%s'.  The correct number of arguments is 5.",
			argc, buffer );
	}

	pv_name = argv[3];
	rf_name = argv[4];

	/* Split the MX record field name into the record name part
	 * and the field name part.
	 */

	dot_ptr = strrchr( argv[4], '.' );

	if ( dot_ptr == NULL ) {
		record_name = argv[4];
		field_name = value_name;
	} else {
		*dot_ptr = '\0';

		record_name = argv[4];
		field_name = dot_ptr + 1;
	}

#if MXCA_EXPORT_DEBUG
	MX_DEBUG(-2,("%s: record_name = '%s', field_name = '%s'",
		fname, record_name, field_name));
#endif

	mx_record = mx_get_record( record_list_head, record_name );

	if ( mx_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NOT_FOUND, fname,
		"Record '%s' not found in the MX database.", record_name );
	}

	mx_field = mx_get_record_field( mx_record, field_name );

	if ( mx_field == (MX_RECORD_FIELD *) NULL ) {
		return mx_error( MXE_NOT_FOUND, fname,
		"Record '%s' does not contain a field called '%s'.",
			record_name, field_name );
	}

	/* If it was not passed to us, then find the mxcaServer object
	 * for this record list.
	 */

	if ( epics_server == (mxcaServer *) NULL ) {
		epics_server = mxca_get_epics_server( record_list_head );

		if ( epics_server == (mxcaServer *) NULL ) {
			return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		    "Cannot find an mxcaServer object for MX record list %p",
				record_list_head );
		}
	}

	/* If the application_ptr for the MX_RECORD contains a pointer to
	 * an mxcaRecord object, then we will want to copy that to the
	 * mxcaPV object that we are just about to create.
	 */

	if ( mx_record->application_ptr != NULL ) {
		epics_record = (mxcaRecord *) mx_record->application_ptr;
	} else {
		epics_record = NULL;
	}

	/* Create the new PV. */

	epics_pv = new mxcaPV( epics_server, epics_record, pv_name );

	if ( epics_pv == (mxcaPV *) NULL ) {
		return mx_error( MXE_UNKNOWN_ERROR, fname,
			"We failed to create EPICS PV '%s', "
			"but did not throw an exception.", pv_name );
	}

	epics_pv->configure_mx_field_pv( mx_field );

	return MX_SUCCESSFUL_RESULT;
}

/*----------------------------------------------------------------------*/

static mx_status_type
mxca_export_epics_motor( mxcaServer *epics_server,
			MX_RECORD *record_list_head,
			char *buffer, int argc, char **argv )
{
	static const char fname[] = "mxca_export_epics_motor()";

	MX_RECORD *current_record;
	mxcaRecord *epics_record;
	int match;
	char *pv_prefix, *wildcard;
	char epics_record_name[80];

	if ( argc != 5 ) {
		return mx_error( MXE_UNPARSEABLE_STRING, fname,
		"Illegal number of arguments (%d) for 'epics_motor' export "
		"statement '%s'.  The correct number of arguments is 5.",
			argc, buffer );
	}

	pv_prefix = argv[3];
	wildcard = argv[4];

#if MXCA_EXPORT_DEBUG
	MX_DEBUG(-2,
	("%s invoked for EPICS PV prefix = '%s', match string = '%s'",
		fname, pv_prefix, wildcard));
#endif

	/* Walk through the record list looking for motor records. */

	current_record = record_list_head->next_record;

	while ( current_record != record_list_head ) {
		if ( current_record->mx_class == MXC_MOTOR ) {
			/* Does the name of this motor match the wildcard? */

			match = mx_match( wildcard, current_record->name );

			if ( match ) {
				snprintf( epics_record_name,
					sizeof(epics_record_name),
					"%s%s", pv_prefix,
						current_record->name );

#if MXCA_EXPORT_DEBUG
				MX_DEBUG(-2,
				("%s: Exporting PVs for motor '%s' as '%s'",
					fname, current_record->name,
					epics_record_name ));
#endif

				epics_record = new mxcaMotorRecord(
							epics_server,
							epics_record_name,
							current_record );
			}
		}
		current_record = current_record->next_record;
	}

	return MX_SUCCESSFUL_RESULT;
}

/*----------------------------------------------------------------------*/

static mx_status_type
mxca_export_epics_scaler( mxcaServer *epics_server,
			MX_RECORD *record_list,
			char *buffer, int argc, char **argv )
{
	static const char fname[] = "mxca_export_epics_scaler()";

	mxcaScalerRecord *epics_scaler;
	MX_RECORD *mx_timer_record;
	MX_RECORD *mx_input_record;
	char *epics_scaler_name, *mx_timer_name, *mx_input_name;
	int i, num_inputs;

	if ( argc < 6 ) {
		return mx_error( MXE_UNPARSEABLE_STRING, fname,
		"Insufficient arguments (%d) for 'epics_scaler' export "
		"statement '%s'.  There should be at least 6 arguments.",
			argc, buffer );
	}

	epics_scaler_name = argv[3];
	mx_timer_name     = argv[4];
	num_inputs        = atoi(argv[5]);

#if MXCA_EXPORT_DEBUG
	MX_DEBUG(-2,("%s invoked for EPICS scaler '%s'.  buffer = '%s'",
		fname, epics_scaler_name, buffer));
#endif

	mx_timer_record = mx_get_record( record_list, mx_timer_name );

	if ( mx_timer_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NOT_FOUND, fname,
		"MX timer '%s' was not found in the MX database.",
			mx_timer_name );
	}

	/* Create the EPICS scaler record. */

	epics_scaler = new mxcaScalerRecord( epics_server,
						epics_scaler_name,
						mx_timer_record );

	if ( epics_scaler == (mxcaRecord *) NULL ) {
		return mx_error( MXE_UNKNOWN_ERROR, fname,
		"The attempt to create EPICS scaler '%s' failed, "
		"but no exception was thrown.", epics_scaler_name );
	}

	/* Add the input channels to the scaler record. */

	for ( i = 6; i < argc; i++ ) {
		mx_input_name = argv[i];

		mx_input_record = mx_get_record( record_list, mx_input_name );

		if ( mx_input_record == (MX_RECORD *) NULL ) {
			(void) mx_error( MXE_NOT_FOUND, fname,
			"MX input device '%s' was not found.  Skipping.",
				mx_input_name );
		} else {
			(void) epics_scaler->add_input( mx_input_record );
		}
	}

	epics_scaler->update_number_of_channels();

	return MX_SUCCESSFUL_RESULT;
}

/*----------------------------------------------------------------------*/

static mx_status_type
mxca_export_callback( MX_RECORD *record_list, char *buffer )
{
	static const char fname[] = "mxca_export_callback()";

	mxcaServer *epics_server;
	char *dup_buffer;
	int argc;
	char **argv;
	mx_status_type mx_status;

	mx_status = MX_SUCCESSFUL_RESULT;

	/* Make sure that this is the list head record. */

	record_list = record_list->list_head;

	epics_server = mxca_get_epics_server( record_list );

#if MXCA_EXPORT_DEBUG
	MX_DEBUG(-2,("%s invoked for buffer '%s'.", fname, buffer));
#endif

	dup_buffer = strdup( buffer );

	mx_string_split( dup_buffer, " ", &argc, &argv );

	if ( argc < 3 ) {
		mx_free(argv); mx_free(dup_buffer);

		return mx_error( MXE_UNPARSEABLE_STRING, fname,
		"An '!export epics' statement should contain at least "
		"3 tokens.  The statement '%s' only contains %d tokens.",
			buffer, argc );
	}

	if ( strcmp( argv[2], "epics_pv" ) == 0 ) {
		mx_status = mxca_export_epics_pv( epics_server, record_list, 
							buffer, argc, argv );
	} else
	if ( strcmp( argv[2], "epics_motor" ) == 0 ) {
		mx_status = mxca_export_epics_motor( epics_server, record_list, 
							buffer, argc, argv );
	} else
	if ( strcmp( argv[2], "epics_scaler" ) == 0 ) {
		mx_status = mxca_export_epics_scaler( epics_server,record_list, 
							buffer, argc, argv );
	} else {
		mx_status = mx_error( MXE_UNPARSEABLE_STRING, fname,
		"The export type '%s' for EPICS export statement '%s' "
		"does not match any known export type.",
			argv[2], buffer );
	}

	mx_free(argv); mx_free(dup_buffer);

	return mx_status;
}

/*----------------------------------------------------------------------*/

/* mxca_export_callback_shim() is a function with C linkage that acts as
 * a front end for the C++ mxca_export_callback() function.  It exists
 * because the MX function mx_register_export_callback() expects to receive
 * a C function pointer.
 */

extern "C" {
MX_EXPORT mx_status_type
mxca_export_callback_shim( MX_RECORD *record_list, char *buffer )
{
	mx_status_type mx_status;

	mx_status = mxca_export_callback( record_list, buffer );

	return mx_status;
}
}  /* End of extern "C" */

